@@ -31,6 +31,7 @@
 #include "ykcore_lcl.h"
 #include "ykcore_backend.h"
 #include "yktsd.h"
+#include "ykbzero.h"
 
 /* To get modhex and crc16 */
 #include <yubikey.h>
@@ -200,6 +201,12 @@ int yk_get_capabilities(YK_KEY *yk, uint8_t slot, unsigned int flags,
 	response_len = capabilities[0];
 	response_len++;
 
+	/* validate the length we got back from the hardware */
+	if (response_len > *len) {
+		yk_errno = YK_EWRONGSIZ;
+		return 0;
+	}
+
 	*len = response_len;
 	return 1;
 }
@@ -253,6 +260,7 @@ int yk_write_device_info(YK_KEY *yk, unsigned char *buf, unsigned int len)
 int yk_write_command(YK_KEY *yk, YK_CONFIG *cfg, uint8_t command,
 		    unsigned char *acc_code)
 {
+	int ret;
 	unsigned char buf[sizeof(YK_CONFIG) + ACC_CODE_SIZE];
 
 	/* Update checksum and insert config block in buffer if present */
@@ -271,8 +279,9 @@ int yk_write_command(YK_KEY *yk, YK_CONFIG *cfg, uint8_t command,
 	if (acc_code)
 		memcpy(buf + sizeof(YK_CONFIG), acc_code, ACC_CODE_SIZE);
 
-	return _yk_write(yk, command, buf, sizeof(buf));
-
+	ret = _yk_write(yk, command, buf, sizeof(buf));
+	insecure_memzero(buf, sizeof(buf));
+	return ret;
 }
 
 int yk_write_config(YK_KEY *yk, YK_CONFIG *cfg, int confnum,
@@ -380,6 +389,7 @@ int yk_challenge_response(YK_KEY *yk, uint8_t yk_cmd, int may_block,
 				&bytes_read)) {
 		return 0;
 	}
+
 	return 1;
 }
 
@@ -608,11 +618,22 @@ int yk_read_response_from_key(YK_KEY *yk, uint8_t slot, unsigned int flags,
 			if ((data[FEATURE_RPT_SIZE - 1] & 31) == 0) {
 				if (expect_bytes > 0) {
 					/* Size of response is known. Verify CRC. */
-					int crc = yubikey_crc16(buf, expect_bytes + 2);
+					expect_bytes += 2;
+					int crc = yubikey_crc16(buf, expect_bytes);
 					if (crc != YK_CRC_OK_RESIDUAL) {
 						yk_errno = YK_ECHECKSUM;
 						return 0;
 					}
+
+					/* since we get data in chunks of 7 we need to round expect bytes out to the closest higher multiple of 7 */
+					if(expect_bytes % 7 != 0) {
+						expect_bytes += 7 - (expect_bytes % 7);
+					}
+
+					if (*bytes_read != expect_bytes) {
+						yk_errno = YK_EWRONGSIZ;
+						return 0;
+					}
 				}
 
 				/* Reset read mode of Yubikey before returning. */
@@ -648,6 +669,7 @@ int yk_write_to_key(YK_KEY *yk, uint8_t slot, const void *buf, int bufcount)
 	YK_FRAME frame;
 	unsigned char repbuf[FEATURE_RPT_SIZE];
 	int i, seq;
+	int ret = 0;
 	unsigned char *ptr, *end;
 
 	if (bufcount > sizeof(frame.payload)) {
@@ -697,16 +719,20 @@ int yk_write_to_key(YK_KEY *yk, uint8_t slot, const void *buf, int bufcount)
 		 */
 		if (! yk_wait_for_key_status(yk, slot, 0, WAIT_FOR_WRITE_FLAG,
 					     false, SLOT_WRITE_FLAG, NULL))
-			return 0;
+			goto end;
 #ifdef YK_DEBUG
 		_yk_hexdump(repbuf, FEATURE_RPT_SIZE);
 #endif
 		if (!_ykusb_write(yk, REPORT_TYPE_FEATURE, 0,
 				  (char *)repbuf, FEATURE_RPT_SIZE))
-			return 0;
+			goto end;
 	}
 
-	return 1;
+	ret = 1;
+end:
+	insecure_memzero(&frame, sizeof(YK_FRAME));
+	insecure_memzero(repbuf, sizeof(repbuf));
+	return ret;
 }
 
 int yk_force_key_update(YK_KEY *yk)
